/* $Id: pca200e.c,v 1.6 1997/04/17 14:47:10 ud3 Exp $ */

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/bios32.h>
#include <linux/pci.h>
#include <linux/malloc.h>
#include <linux/atmdev.h>
#include <linux/sonet.h>
#include <asm/io.h>
#include <asm/string.h>

#ifdef MODULE
#include <linux/module.h>
#endif

#include "pca200e.h"
#include "pca200e_firmw.h"

#ifdef CONFIG_ATM_PCA200E_DEBUG
#define DPRINTK(format,args...) printk(format,##args)
#else
#define DPRINTK(format,args...)
#endif


/* support old naming sillyness */
#if LINUX_VERSION_CODE < 0x020100                                     
#define ioremap vremap
#define iounmap vfree                                                     
#endif




struct pca200e_dev *	pca200e_boards		= NULL;
static const char	pca200e_firmware[]	= PCA200E_FIRMWARE;


int
pca200e_card_detect(struct pca200e_dev * pca200e, unsigned index)
{
  unsigned short pci_command;
  
  if (!pcibios_present())
    {
      printk("PCA-200E driver but no PCI BIOS ?\n");
      return -ENODEV;
    };
  
  if(pcibios_find_device(PCI_VENDOR_ID_FORE,
			 PCI_DEVICE_ID_FORE_PCA200E,
                         index,
			 &pca200e->bus,
			 &pca200e->dev_fn))
    return -ENODEV;
  
  if (pcibios_read_config_dword(pca200e->bus,
				pca200e->dev_fn,
				PCI_BASE_ADDRESS_0,
                                (unsigned*) &pca200e->phys_rambase) != PCIBIOS_SUCCESSFUL)
    {
      printk("Can't read membase\n");
      return -EIO;
    };
  
  if (pcibios_read_config_byte(pca200e->bus,
			       pca200e->dev_fn,
			       PCI_INTERRUPT_LINE,
                               (unsigned char*) &pca200e->irq) != PCIBIOS_SUCCESSFUL)
    {
      printk("Can't read IRQ\n");
      return -EIO;
    };

   if (pcibios_write_config_byte(pca200e->bus,
				pca200e->dev_fn,
				PCI_LATENCY_TIMER, 
				255) != PCIBIOS_SUCCESSFUL)
    {
      printk("Can't write PCI_LATENCY_TIMER\n");
      return -EIO;
    };
  
  
  if (pcibios_read_config_word(pca200e->bus,
			       pca200e->dev_fn,
			       PCI_COMMAND,
                               &pci_command) != PCIBIOS_SUCCESSFUL)
    {
      printk("Can't read PCI_COMMAND\n");
      return -EIO;
    };
  
  pci_command |= (PCI_COMMAND_MEMORY
		  | PCI_COMMAND_SPECIAL
		  | PCI_COMMAND_MASTER);

  pci_command &= ~(PCI_COMMAND_IO);

  if (pcibios_write_config_word(pca200e->bus,
				pca200e->dev_fn,
				PCI_COMMAND, 
				pci_command) != PCIBIOS_SUCCESSFUL)
    {
      printk("Can't write PCI_COMMAND\n");
      return -EIO;
    };
  
  return 0;
};



int
pca200e_map(struct pca200e_dev* pca200e)
{
  pca200e->virt_rambase = (void*) ioremap((unsigned long) pca200e->phys_rambase,
					  PCA200E_HARDWARE_BASE_LENGTH);
  if (!pca200e->virt_rambase)
    {
      printk(DEV_LABEL "ioremap: can't set up mapping\n");
      return -EFAULT;
    }
  pca200e->cardstate = cs_map;
  return 0;
};

int
pca200e_reset(struct pca200e_dev* pca200e)
{
  int i;
  monitor_t* m   = (monitor_t*)(pca200e->virt_rambase + PCA200E_MON960_OFFSET);
  writel(cold_start, &m->bstat);
  writeb(0x01, pca200e->virt_rambase + PCA200E_HCR_OFFSET);
  i=jiffies; while (jiffies <= i+1);	/* hold reset for 10..20 ms	*/
  writeb(0x00, pca200e->virt_rambase + PCA200E_HCR_OFFSET);
  i=jiffies;
  while (jiffies <= i+300)		/* wait max. 3s for selftest	*/
    if (readl(&m->bstat) == self_test_ok)
      {
	pca200e->cardstate = cs_reset;
	return 0;
      };
  printk(DEV_LABEL "self test FAILED.\n");
  return -EIO;
};

void
pca200e_loadfirm(struct pca200e_dev* pca200e)
{
  memcpy_toio(pca200e->virt_rambase + ((struct firmheader*) pca200e_firmware)->loadoffset,
	      pca200e_firmware,
	      sizeof(pca200e_firmware));
  pca200e->cardstate = cs_loadfirm;
};

void
pca200e_mon960put_c(struct pca200e_dev* pca200e, unsigned char c)
{
  monitor_t * m = (monitor_t *)(pca200e->virt_rambase+PCA200E_MON960_OFFSET);
  unsigned long i;
  
  i = jiffies;
  while(jiffies < i+2)	/* 20 ms */
    if (readl(&m->uart.recv) & PCA200E_MON960_UART_AVAIL)
      {
	writel(PCA200E_MON960_UART_FREE, &m->uart.recv);
	i=jiffies;
      };
  
  writel(c | PCA200E_MON960_UART_AVAIL, &m->uart.send);
  
  i = jiffies;
  while(jiffies < i+2)	/* 20 ms */
    if (readl(&m->uart.recv) & PCA200E_MON960_UART_AVAIL)
      {
	writel(PCA200E_MON960_UART_FREE, &m->uart.recv);
	i=jiffies;
      };
};

unsigned int
pca200e_startfirm(struct pca200e_dev* pca200e)
{
  monitor_t* m   = (monitor_t*)(pca200e->virt_rambase + PCA200E_MON960_OFFSET);
  char s[16];
  int i;

  sprintf(s,"\rgo %08X\r",
	  ((struct firmheader*) pca200e_firmware)->startoffset);
  for (i=0; s[i] != 0x00; i++)
    pca200e_mon960put_c(pca200e, s[i]);
  i = jiffies;
  while (jiffies < i+300)	/* wait max. 3s for firmware getting started */
    if(readl(&m->bstat) == cp_running)
      {
	pca200e->cardstate = cs_startfirm;
	return 0;
      };
  printk(DEV_LABEL "firmware DIDN'T START.\n");
  return -EIO;
};


unsigned int
pca200e_initialize(struct pca200e_dev* pca200e)
{
  int i,j;

  pca200e->cpqueue = (cp_queues_t*) (pca200e->virt_rambase + PCA200E_COMMON_ORIGIN);

  writel(0,			&pca200e->cpqueue->init.receive_threshold);
  writel(NUM_CONNECTIONS,	&pca200e->cpqueue->init.num_connect);
  writel(QSIZE_CMD,		&pca200e->cpqueue->init.cqueue_len);
  writel(QSIZE_RX,		&pca200e->cpqueue->init.rqueue_len);
  writel(QSIZE_TX,		&pca200e->cpqueue->init.tqueue_len);
  writel(0xffffffff,		&pca200e->cpqueue->init.conless_vpvc);
  writel(RPD_EXTENSION,		&pca200e->cpqueue->init.rpd_extension);
  writel(TPD_EXTENSION,		&pca200e->cpqueue->init.tpd_extension);

  writel(QSIZE_RESUPPLY,	&pca200e->cpqueue->init.small_b1.queue_length);
  writel(SMALL_B1_BUFF_SIZE,	&pca200e->cpqueue->init.small_b1.buffer_size);
  writel(SMALL_B1_POOL_SIZE,	&pca200e->cpqueue->init.small_b1.pool_size);
  writel(RBD_BLK_SIZE,		&pca200e->cpqueue->init.small_b1.supply_blksize);

  writel(QSIZE_RESUPPLY,	&pca200e->cpqueue->init.large_b1.queue_length);
  writel(LARGE_B1_BUFF_SIZE,	&pca200e->cpqueue->init.large_b1.buffer_size);
  writel(LARGE_B1_POOL_SIZE,	&pca200e->cpqueue->init.large_b1.pool_size);
  writel(RBD_BLK_SIZE,		&pca200e->cpqueue->init.large_b1.supply_blksize);

  writel(0, &pca200e->cpqueue->init.small_b2.queue_length);
  writel(0, &pca200e->cpqueue->init.small_b2.buffer_size);
  writel(0, &pca200e->cpqueue->init.small_b2.pool_size);
  writel(0, &pca200e->cpqueue->init.small_b2.supply_blksize);

  writel(0, &pca200e->cpqueue->init.large_b2.queue_length);
  writel(0, &pca200e->cpqueue->init.large_b2.buffer_size);
  writel(0, &pca200e->cpqueue->init.large_b2.pool_size);
  writel(0, &pca200e->cpqueue->init.large_b2.supply_blksize);

  writel(stat_pending,	&pca200e->cpqueue->init.status);
  writel(op_initialize,	&pca200e->cpqueue->init.op);

  i = jiffies; j = readl(&pca200e->cpqueue->heartbeat);
  /* wait max. 20ms for command completion, heartbeat counting upwards or timeout */
  while (jiffies < i+20)
    {
      if((j < readl(&pca200e->cpqueue->heartbeat))
	 || (readl(&pca200e->cpqueue->init.status) & stat_complete))
	break;
      else
	j = readl(&pca200e->cpqueue->heartbeat);
    };

  /*
   * TODO: status still != complete while heartbeat is already counting
   *       upwards will lead to abort although all went OK
   */ 
  if (readl(&pca200e->cpqueue->init.status) != stat_complete)
    {
      printk(DEV_LABEL "init command failed!!!\n");
      return -EIO;
    };
  pca200e->cardstate = cs_init;
  return 0;
};

int
pca200e_initcmdqueue(struct pca200e_dev* pca200e)
{
  int i;
  qhost_cmd_t *tmp1, *tmp2 = NULL;
  
  pca200e->cmd_status_page = (void*) __get_dma_pages(GFP_KERNEL,
						     (QSIZE_CMD*32+4095)/4096);
  if (!(pca200e->cmd_status_page))
    {
      printk(DEV_LABEL "initcmdqueue: cmd_status_page: get_dma_pages failed\n");
      return -ENOMEM;
    };
  for (i=0; i<QSIZE_CMD; i++)
    {
      tmp1 = kmalloc(sizeof(qhost_cmd_t), GFP_KERNEL);
      if (!tmp1)
	{
	  printk(DEV_LABEL "initcmdqueue: kmalloc failed - index=%d\n",i);
	  return -ENOMEM;
	};
      memset((void*) tmp1, 0, sizeof(qhost_cmd_t));
      if(i)
	tmp2->next = tmp1;
      else
	pca200e->cmd_queue = tmp1;

      tmp1->qentry = (qcard_cmd_t*) ((unsigned)(pca200e->virt_rambase)
				     +(unsigned)(pca200e->cpqueue->command_queue)
				     +i*sizeof(qcard_cmd_t));
      tmp1->status = (fw_status_t*)((unsigned)(pca200e->cmd_status_page)+i*32);
      writel((unsigned long) tmp1->status, &tmp1->qentry->stataddr);
      tmp1->status[0] = stat_free;
      tmp2 = tmp1;
    };
  tmp1->next=pca200e->cmd_queue;

  pca200e->cardstate = cs_init_cmd;
  return 0;
};

int
pca200e_inittxqueue(struct pca200e_dev* pca200e)
{
  int i;
  qhost_tx_t *tmp1, *tmp2=0;
  
  pca200e->tpd_page = (void*) __get_dma_pages(GFP_KERNEL,
					      (QSIZE_TX*TPD_SIZE+4095)/4096);
  if (!pca200e->tpd_page)
    {
      printk(DEV_LABEL "inittxqueue: tpd_page: get_dma_pages failed\n");
      return -ENOMEM;
    };
  pca200e->tx_status_page = (void*) __get_dma_pages(GFP_KERNEL,
						    (QSIZE_TX*32+4095)/4096);
  if (!pca200e->tx_status_page)
    {
      printk(DEV_LABEL "inittxqueue: tx_status_page: get_dma_pages failed\n");
      return -ENOMEM;
    };
  for (i=0; i<QSIZE_TX; i++)
    {
      tmp1 = kmalloc(sizeof(qhost_tx_t), GFP_KERNEL);
      if (!tmp1)
	{
	  printk(DEV_LABEL "inittxqueue: kmalloc failed - index=%d\n",i);
	  return -ENOMEM;
	};
      memset((void*) tmp1, 0, sizeof(qhost_tx_t));
      if (i)
	tmp2->next = tmp1;
      else
	pca200e->tx_queue = tmp1;

      tmp1->qentry = (qcard_t*) ((unsigned)(pca200e->virt_rambase)
				 +(unsigned)(pca200e->cpqueue->transmit_queue)
				 +i*sizeof(qcard_t));
      tmp1->status = (fw_status_t*)((unsigned)(pca200e->tx_status_page)+i*32);
      writel((unsigned long) tmp1->status, &tmp1->qentry->stataddr);
      tmp1->status[0] = stat_free;
      tmp1->tpd	= (tpd_t*)((unsigned)(pca200e->tpd_page)+i*TPD_SIZE);
      tmp2 = tmp1;
    };
  tmp1->next = pca200e->tx_queue;
  pca200e->tx_queue_tail = pca200e->tx_queue;

  pca200e->cardstate = cs_init_tx;
  return 0;
};

int
pca200e_initrxqueue(struct pca200e_dev* pca200e)
{
  int i;
  qhost_rx_t *tmp1, *tmp2 = NULL;
  
  pca200e->rpd_page = (void*) __get_dma_pages(GFP_KERNEL,
					      (QSIZE_RX*RPD_SIZE+4095)/4096);
  if (!pca200e->rpd_page)
    {
      printk(DEV_LABEL "initrxqueue: rpd_page: get_dma_pages failed\n");
      return -ENOMEM;
    };
  pca200e->rx_status_page = (void*) __get_dma_pages(GFP_KERNEL,
						    (32*QSIZE_RX+4095)/4096);
  if (!pca200e->rx_status_page)
    {
      printk(DEV_LABEL "initrxqueue: rx_status: get_dma_pages failed\n");
      return -ENOMEM;
    };
  for (i=0; i<QSIZE_RX; i++)
    {
      tmp1 = kmalloc(sizeof(qhost_rx_t), GFP_KERNEL);
      if (!tmp1)
	{
	  printk(DEV_LABEL "initrxqueue: kmalloc failed - index=%d\n",i);
	  return -ENOMEM;
	};
      memset((void*) tmp1, 0, sizeof(qhost_rx_t));
      if(i)
	tmp2->next = tmp1;
      else
	pca200e->rx_queue = tmp1;
      
      tmp1->qentry = (qcard_t*) ((unsigned)(pca200e->virt_rambase)
				 +(unsigned)(pca200e->cpqueue->receive_queue)
				 + i*sizeof(qcard_t));
      tmp1->status = (fw_status_t*) ((unsigned)(pca200e->rx_status_page)
				     +i*32);
      writel((unsigned long) tmp1->status, &tmp1->qentry->stataddr);
      tmp1->status[0] = stat_free;
      tmp1->rpd	= (rpd_t*)((unsigned)(pca200e->rpd_page)
			   +i*((sizeof(rpd_t)+31)/32)*32);
      tmp1->qentry->ioblock = (void*) tmp1->rpd;
      tmp2 = tmp1;
    };
  tmp1->next = pca200e->rx_queue;
  
  pca200e->cardstate = cs_init_rx;
  return 0;
};

int
pca200e_buildrbuffqueue(struct pca200e_dev* pca200e, int sml)
{
  unsigned i;
  qhost_rbuff_t* tmp1 = NULL;
  qhost_rbuff_t* tmp2 = NULL;
  
  qhost_rbuff_t** ptr;		/* pointer to queue pointer	*/
  unsigned int count;		/* number of buffers		*/
  unsigned int size;		/* buffer size			*/
  buffer_magn_t mag;		/* buffer magnitude		*/
  
  ptr = (sml ? &pca200e->smlb1_rbuff_head : &pca200e->lrgb1_rbuff_head);
  count = (sml ? SMALL_B1_NUM_RBUFFS : LARGE_B1_NUM_RBUFFS);
  size = (sml ? SMALL_B1_BUFF_SIZE : LARGE_B1_BUFF_SIZE);
  mag = (sml ? bmag_small : bmag_large);
  
  for (i=0; i< count; i++)
    {
      tmp1 = kmalloc(sizeof(qhost_rbuff_t), GFP_KERNEL);
      if (!tmp1)
	{
	  printk(DEV_LABEL "buildrbuffqueue: kmalloc failed - index=%d\n",i);
	  *ptr = tmp2;
	  return -ENOMEM;
	};
      pca200e->bufferlist[((unsigned int) (pca200e->bufferlist[0]))++] = tmp1;
      memset((void*) tmp1, 0, sizeof(qhost_rbuff_t));
      tmp1->type = mag;    
      tmp1->next = tmp2;
      tmp1->buff = kmalloc(size, GFP_KERNEL);
      if (!tmp1->buff)
	{
	  printk(DEV_LABEL "buildrbuffqueue: buffer kmalloc failed - index=%d\n",i);
	  *ptr = tmp1;
	  return -ENOMEM;
	};
      tmp2 = tmp1;
    };
  *ptr = tmp1;
  return 0;
};

void
pca200e_supply(struct pca200e_dev* pca200e, int sml)
{
  int i;
  
  qhost_supp_t**  sptr;
  qhost_rbuff_t** rptr;
  int* cptr;

  sptr = (sml ? &pca200e->smlb1_queue : &pca200e->lrgb1_queue);
  rptr = (sml ? &pca200e->smlb1_rbuff_head : &pca200e->lrgb1_rbuff_head);
  cptr = (sml ? &pca200e->smlb1_count : &pca200e->lrgb1_count);

  for (i=0; i<RBD_BLK_SIZE; i++)
    {
      (*sptr)->rbd[i].buffer = (*rptr)->buff;
      (*sptr)->rbd[i].handle = (void*) *rptr;
      *rptr = (*rptr)->next;
    };

  (*sptr)->status[0] = stat_pending;
  writel((unsigned long) (*sptr)->rbd, &((*sptr)->qentry->ioblock));
  *sptr = (*sptr)->next;
  *cptr += RBD_BLK_SIZE;
};

int
pca200e_initbuffqueue(struct pca200e_dev* pca200e, int sml)
{
  int i;
  qhost_supp_t *tmp1, *tmp2 = NULL;
  void** stpptr;
  rbd_t** rbpptr;
  qhost_supp_t** qptr;
  qcard_t* cpq;
  int* cptr;
  enum cardstate_t state;

  stpptr = (sml ? &pca200e->smlb1_status_page : &pca200e->lrgb1_status_page);
  rbpptr = (sml ? &pca200e->smlb1_rbd_page : &pca200e->lrgb1_rbd_page);
  qptr = (sml ? &pca200e->smlb1_queue : &pca200e->lrgb1_queue);
  cpq = (sml ? pca200e->cpqueue->small_b1_queue : pca200e->cpqueue->large_b1_queue);
  cptr = (sml ? &pca200e->smlb1_count : &pca200e->lrgb1_count);
  state = (sml ? cs_init_sb1 : cs_init_lb1);

  *stpptr = (void*) __get_dma_pages(GFP_KERNEL, (32*QSIZE_RESUPPLY+4095)/4096);
  if (!*stpptr)
    {
      printk(DEV_LABEL "initbuffqueue: status_page: get_dma_pages failed\n");
      return -ENOMEM;
    };
  *rbpptr = (void*) __get_dma_pages(GFP_KERNEL, (QSIZE_RESUPPLY*RBD_BLK_SIZE*sizeof(rbd_t)+4095)/4096);
  if (!*rbpptr)
    {
      printk(DEV_LABEL "initbuffqueue: rbd_page: get_dma_pages failed\n");
      return -ENOMEM;
    };
  for (i=0; i<QSIZE_RESUPPLY; i++)
    {
      tmp1 = kmalloc(sizeof(qhost_supp_t), GFP_KERNEL);
      if (!tmp1)
	{
	  printk(DEV_LABEL "initbuffqueue: kmalloc failed - index=%d\n",i);
	  return -ENOMEM;
	};
      memset((void*) tmp1, 0, sizeof(qhost_supp_t));
      if (i)
	tmp2->next = tmp1;
      else
	*qptr = tmp1;
      
      tmp1->qentry = (qcard_t*) ((unsigned)(pca200e->virt_rambase)
				 +(unsigned) cpq
				 +i*sizeof(qcard_t));
      tmp1->status = (fw_status_t*)((unsigned)(*stpptr)
				    +i*32);
      writel((unsigned long) tmp1->status, &tmp1->qentry->stataddr);
      tmp1->status[0] = stat_free;
      tmp1->rbd = (rbd_t*)((unsigned)(*rbpptr)
			   +i*RBD_BLK_SIZE*sizeof(rbd_t));
      tmp2 = tmp1;
    };
  tmp1->next = *qptr;
  *cptr = 0;
  
  if (pca200e_buildrbuffqueue(pca200e, sml))
    return -ENOMEM;
  pca200e_supply(pca200e, sml);
  
  pca200e->cardstate = state;
  return 0;
};

void
pca200e_readprom(struct pca200e_dev* pca200e)
{
  int i;
  expansion_rom_data_t* exp;

  exp = (expansion_rom_data_t*) __get_dma_pages(GFP_KERNEL,1);
  if (!exp)
    {
      return;
    };
  memset((void*) exp, 0, sizeof(expansion_rom_data_t));
  writel((unsigned long) exp,
	 &pca200e->cmd_queue->qentry->cmd.prom_data_param.prom_data_buff);
  pca200e->cmd_queue->status[0] = stat_pending;
  writel(op_get_prom_data,
	 &pca200e->cmd_queue->qentry->cmd.prom_data_param.op);
  
  i = jiffies; while (jiffies < i+30)
    if(pca200e->cmd_queue->status[0] == stat_complete)
      break;
  if (!(pca200e->cmd_queue->status[0] == stat_complete))
    printk(DEV_LABEL "readprom command failed!!!\n");
  
  printk(DEV_LABEL "Rev %c, S/N: %d, ESI: %02x:%02x:%02x:%02x:%02x:%02x\n",
	 (char)((exp->hw_version & 0xff)+'@'),
	 exp->serial_number,
	 exp->mac_addr[2], exp->mac_addr[3], exp->mac_addr[4],
	 exp->mac_addr[5], exp->mac_addr[6], exp->mac_addr[7]);
  
  for (i=2;i<8;i++)
    pca200e->mac[i-2] = exp->mac_addr[i];
  
  pca200e->cmd_queue->status[0] = stat_free;
  pca200e->cmd_queue = pca200e->cmd_queue->next;
  free_pages((long)exp,1);
};

void
pca200e_getstats(struct pca200e_dev* pca200e, stats200_t* stats)
{
  int i;
  qhost_cmd_t*	cq;

  cq = pca200e->cmd_queue;
  writel((unsigned int) stats, &cq->qentry->cmd.stats_param.stats_buff);
  cq->status[0] = stat_pending;
  writel(op_request_stats, &cq->qentry->cmd.stats_param.op);
  
  i = jiffies;
  while (jiffies < i+40)
    if(cq->status[0] == stat_complete)
      break;
  
  if (!(cq->status[0] == stat_complete))
    printk(DEV_LABEL "getstats command failed!!!\n");
  
  cq->status[0] = stat_free;
  pca200e->cmd_queue = cq->next;
};

void
pca200e_inttx(struct atm_dev* dev)
{
  qhost_tx_t*	txq;
  int	i;
  
  txq = PCA200E_DEV(dev)->tx_queue_tail;
  for (i=0; i<QSIZE_TX; i++)
    {
      if (txq->status[0] & stat_complete)
	{
	  txq->vcc->pop(txq->vcc, txq->skb);
	  if (txq->status[0] & stat_error)
	    txq->vcc->stats->tx_err++;
	  else
	    txq->vcc->stats->tx++;
	  txq->status[0] = stat_free;
	}
      else
	break;
      txq = txq->next;
    };
  PCA200E_DEV(dev)->tx_queue_tail = txq;
};

void
pca200e_intrx(struct atm_dev* dev)
{
  qhost_rx_t*	rxq;
  struct pca200e_dev* pca200e;
  int	i;
  
  pca200e = PCA200E_DEV(dev);
  rxq = pca200e->rx_queue;
  for (i=0; i<QSIZE_RX; i++)
    {
      if (rxq->status[0] & stat_complete)
	{
	  rpd_t* rpd = rxq->rpd;
	  struct atm_vcc* vcc;
	  int j;

	  for (vcc = dev->vccs; vcc; vcc = vcc->next)
	    if (((vcc->vpi) == (rpd->atm_header.vpi)) &&
		((vcc->vci) == (rpd->atm_header.vci)))
	      break;
	  if (vcc)
	    {
	      struct sk_buff* skb;
	      int pdulen = 0;
	      
	      for (j=0; j<rpd->nseg; j++)
		  pdulen += rpd->segment[j].length;
	      
	      /* pca200e->rx_queue = rxq;  for fetch */
	      skb = vcc->peek(vcc, pdulen, NULL);
	      
	      if (skb)
		{
		  skb->free = 1;
		  
		  for (j=0; j<rpd->nseg; j++)
		    memcpy(skb_put(skb, rpd->segment[j].length),
			   (rpd->segment[j].handle->buff),
			   rpd->segment[j].length);
		  
		  vcc->push(vcc, skb);
		  vcc->stats->rx++;
		}
	      else
		vcc->stats->rx_drop++;
	    };
	  
	  for (j=0; j<rpd->nseg; j++)
	    {
	      switch (((qhost_rbuff_t*)(rpd->segment[j].handle))->type)
		{
		case bmag_small: 
		  {
		    pca200e->smlb1_count--;
		    (qhost_rbuff_t*) (rpd->segment[j].handle)->next = pca200e->smlb1_rbuff_head;
		    pca200e->smlb1_rbuff_head = (qhost_rbuff_t*) (rpd->segment[j].handle);
		  }; break;
		case bmag_large: 
		  {
		    pca200e->lrgb1_count--;
		    (qhost_rbuff_t*) (rpd->segment[j].handle)->next = pca200e->lrgb1_rbuff_head;
		    pca200e->lrgb1_rbuff_head = (qhost_rbuff_t*) (rpd->segment[j].handle);
		  }; break;
		};
	    };
	  if (pca200e->smlb1_count <= SMALL_B1_LOW_WATER)
	    pca200e_supply(pca200e, 1);
	  if (pca200e->lrgb1_count <= LARGE_B1_LOW_WATER)
	    pca200e_supply(pca200e, 0);
	  writel((unsigned long) rxq->rpd, &rxq->qentry->ioblock);
	  rxq->status[0] = stat_pending;
	}
      else
	break;
      rxq = rxq->next;
    };
  pca200e->rx_queue = rxq;
};


static int
pca200e_send(struct atm_vcc *vcc, struct sk_buff *skb)
{
  qhost_tx_t*	txq;
  tpd_t*	tpd;

  txq = PCA200E_DEV(vcc->dev)->tx_queue;
  if (txq->status[0] != stat_free)
    {
      pca200e_inttx(vcc->dev);	/* try to free completed txqueue entries */
      if (txq->status[0] != stat_free)
	{
	  vcc->stats->tx_err++;
	  printk(DEV_LABEL "TX problem: itf=%d, queue saturated\n", vcc->itf);
	  return -EBUSY;
	};
    };

  txq->vcc = vcc;
  txq->skb = skb;

  tpd = txq->tpd;
  tpd->rc_stream_desc.data_cells = 0x0000;
  tpd->rc_stream_desc.empty_cells = 0x0000;

  tpd->atm_header.clp = 0;
  tpd->atm_header.plt = 0;
  tpd->atm_header.vci = vcc->vci;
  tpd->atm_header.vpi = vcc->vpi;
  tpd->atm_header.gfc = 0;

  tpd->spec.pdu_len = skb->len;
  tpd->spec.num_seg = 1;
  tpd->spec.aal	    = vcc->aal;
  tpd->spec.intr    = 1;

  tpd->segment[0].buffer = skb->data;
  tpd->segment[0].length = ((skb->len + 3) & 0x1fffc);

  txq->status[0] = stat_pending;
  writel((unsigned long) tpd+TPD_EXTENSION+1, &txq->qentry->ioblock);

  PCA200E_DEV(vcc->dev)->tx_queue = txq->next;
  return 0;
};


int
pca200e_activatevcin(struct pca200e_dev* pca200e,
		     unsigned int activate,
		     unsigned int aal, unsigned int vpi, unsigned int vci,
		     unsigned int mtu)
     /* tries to [de]activate the specified aal/vpi/vci
	activate == TRUE  -> activate
	activate == FALSE -> deactivate
	returns TRUE on success */
{
  int i;
  qhost_cmd_t*	cq;
  
  cq = pca200e->cmd_queue;
  /* vpi has to be zero */
  writel((unsigned int) vci, &cq->qentry->cmd.activate_param.vpvc);
  if (activate && (0 == aal))
    writel((unsigned int) mtu, &cq->qentry->cmd.activate_param.mtu);
  cq->status[0] = stat_pending;
  writel((activate ? op_activate_vcin : op_deactivate_vcin) | (aal << 8),
	 &cq->qentry->cmd.activate_param.op);
  
  i = jiffies;
  while (jiffies < i+40)
    if(cq->status[0] != stat_pending)
      break;
  i = !(cq->status[0] & stat_error);
  
  cq->status[0] = stat_free;
  pca200e->cmd_queue = cq->next;

  return i;
};




void
pca200e_interrupt(int irq, struct atm_dev* dev, struct pt_regs *regs)
{
  /* check if we caused this interrupt */
  if (0x01 != readb(PCA200E_DEV(dev)->virt_rambase + PCA200E_PSR_OFFSET))
    {
      printk(DEV_LABEL "Unexpected Interrupt %d\n", irq);
      return;
    };
  pca200e_intrx(dev);
  pca200e_inttx(dev);
  
  /* signalling interrupt acknowledge */
  writeb(0x10, PCA200E_DEV(dev)->virt_rambase + PCA200E_HCR_OFFSET);
};


static int
pca200e_init(struct atm_dev* dev)
{
  if (!PCA200E_DEV(dev))
    return -EFAULT;
  
  if (pca200e_map(PCA200E_DEV(dev))) return -ENOMEM;
  if (pca200e_reset(PCA200E_DEV(dev))) return -ENOMEM;
  pca200e_loadfirm(PCA200E_DEV(dev));
  if (pca200e_startfirm(PCA200E_DEV(dev))) return -ENOMEM;
  if (pca200e_initialize(PCA200E_DEV(dev))) return -ENOMEM;
  if (pca200e_initcmdqueue(PCA200E_DEV(dev))) return -ENOMEM;
  if (pca200e_inittxqueue(PCA200E_DEV(dev))) return -ENOMEM;
  if (pca200e_initrxqueue(PCA200E_DEV(dev))) return -ENOMEM;
  
  if (pca200e_initbuffqueue(PCA200E_DEV(dev), 1)) return -ENOMEM;
  if (pca200e_initbuffqueue(PCA200E_DEV(dev), 0)) return -ENOMEM;
  
  pca200e_readprom(PCA200E_DEV(dev));
  
  if(request_irq(PCA200E_DEV(dev)->irq,
		 (void*)pca200e_interrupt,
		 SA_INTERRUPT,
		 PCA200E_DEV(dev)->name,
		 dev))
    {
      printk(DEV_LABEL "unable to reserve IRQ %d for %s\n",
	     PCA200E_DEV(dev)->irq, PCA200E_DEV(dev)->name);
      return -EBUSY;
    };
  PCA200E_DEV(dev)->cardstate = cs_int;
  dev->ci_range.vpi_bits = 8;
  dev->ci_range.vci_bits = 10;
  return 0;
};


/*-----------------------------------------------*/
static int
pca200e_open(struct atm_vcc *vcc, short vpi, int vci)
{
  vcc->vpi = vpi;
  vcc->vci = vci;
  vcc->itf = vcc->dev->number;
  DPRINTK(DEV_LABEL "open: %d.%d.%d:%d\n",
	 vcc->itf, vcc->vpi, vcc->vci, vcc->aal);
  if (!pca200e_activatevcin(PCA200E_DEV(vcc->dev),
			    1 /* TRUE */,
			    vcc->aal, vcc->vpi, vcc->vci,
			    0))
    return -EBUSY;
#ifdef MODULE
  MOD_INC_USE_COUNT;
#endif
  vcc->flags |= ATM_VF_ADDR;
  vcc->flags |= ATM_VF_READY;
  return 0;
};


static void
pca200e_close(struct atm_vcc *vcc)
{
  DPRINTK(DEV_LABEL "close: %d.%d.%d:%d\n",
	 vcc->itf, vcc->vpi, vcc->vci, vcc->aal);
  if (pca200e_activatevcin(PCA200E_DEV(vcc->dev),
			   0 /* FALSE */,
			   vcc->aal, vcc->vpi, vcc->vci,
			   0))
    {
#ifdef MODULE
      MOD_DEC_USE_COUNT;
#endif
    };
};


unsigned int
dmaswap(unsigned int in)
{
  asm("bswap %%eax\n"
      : "=a" (in)
      : "a" (in));
  return in;
};

static int
pca200e_ioctl(struct atm_dev *dev,unsigned int cmd, unsigned long arg)
{
  switch (cmd)
    {
    case SONET_GETSTATZ:
    case SONET_GETSTAT:
      {
	struct sonet_stats sost;
	stats200_t* stat;
	
	stat = (stats200_t*) kmalloc(sizeof(stats200_t), GFP_KERNEL);
	if (!stat)
	  return -ENOMEM;
	memset((void*) stat, 0, sizeof(stats200_t));
	pca200e_getstats(PCA200E_DEV(dev), stat);
	
	sost.section_bip = dmaswap(stat->phyoc3.section_bip8_errors);
	sost.line_bip	 = dmaswap(stat->phyoc3.line_bip24_errors);
	sost.path_bip	 = dmaswap(stat->phyoc3.path_bip8_errors);
	sost.line_febe	 = dmaswap(stat->phyoc3.line_febe_errors);
	sost.path_febe	 = dmaswap(stat->phyoc3.path_febe_errors);
	sost.corr_hcs	 = dmaswap(stat->phyoc3.corr_hcs_errors);
	sost.uncorr_hcs	 = dmaswap(stat->phyoc3.ucorr_hcs_errors);
	sost.tx_cells	 = (dmaswap(stat->atm.cells_transmitted)
			    +dmaswap(stat->aal0.cells_transmitted)
			    +dmaswap(stat->aal4.cells_transmitted)
			    +dmaswap(stat->aal5.cells_transmitted));
	sost.rx_cells	 = (dmaswap(stat->atm.cells_received)
			    +dmaswap(stat->aal0.cells_received)
			    +dmaswap(stat->aal4.cells_received)
			    +dmaswap(stat->aal5.cells_received));
	
	memcpy_tofs((struct sonet_stats*)arg, &sost,
		    sizeof(struct sonet_stats));
	kfree((void*)stat);
	break;
      }
    case SONET_SETDIAG:
    case SONET_CLRDIAG:
    case SONET_GETDIAG:
      return 0; break;
    case PCA200E_DUMP:
      {
	stats200_t* stat;
	
	memcpy_tofs((char*)(&(((struct pca200e_dump*) arg)->dev)),
		    (char*)PCA200E_DEV(dev),
		    sizeof(struct pca200e_dev));
	memcpy_tofs((char*)(&(((struct pca200e_dump*) arg)->cpqueue)),
		    (char*)(PCA200E_DEV(dev)->cpqueue),
		    sizeof(cp_queues_t));
	memcpy_tofs((char*)(&(((struct pca200e_dump*) arg)->monitor)),
		    (char*)(PCA200E_DEV(dev)->virt_rambase + PCA200E_MON960_OFFSET),
		    sizeof(monitor_t));

	stat = (stats200_t*) kmalloc(sizeof(stats200_t), GFP_KERNEL);
	if (stat)
	  {
	    pca200e_getstats(PCA200E_DEV(dev), stat);
	    memcpy_tofs((void*) &(((struct pca200e_dump*) arg)->stats),
			(void*) stat,
			sizeof(stats200_t));
	    kfree((void*)stat);
	  };
	return 0;
      };
    default:
      DPRINTK(">pca200e_ioctl: cmd=%08x, arg=%08lx\n",
	      cmd, arg);
    };
  return 0;
};


static const struct atmdev_ops ops =
{
  pca200e_open,
  pca200e_close,
  pca200e_ioctl,
  NULL, /*pca200e_getsockopt,*/
  NULL, /*pca200e_setsockopt,*/
  pca200e_send,
  NULL, /*pca200e_sg_send,*/
  NULL,	/*pca200e_poll,*/
  NULL, /*pca200e_send_oam,*/
  NULL, /*pca200e_phy_put,*/
  NULL, /*pca200e_phy_get,*/
  NULL, /*pca200e_feedback,*/
  NULL, /*pca200e_change_qos,*/
  NULL  /*pca200e_free_rx_skb*/
};


int
pca200e_detect(void)
{
  struct atm_dev*	dev;
  struct pca200e_dev*	pca200e;
  int			index;
  int			i;
  
  pca200e = (struct pca200e_dev *) kmalloc(sizeof(struct pca200e_dev),
					   GFP_KERNEL);
  if (!pca200e)
    {
      return -ENOMEM;
    };
  memset(pca200e, 0, sizeof(struct pca200e_dev));

  index = 0;
  while (!pca200e_card_detect(pca200e, index))
    {
      (unsigned int) pca200e->bufferlist[0] = 1;
      sprintf(pca200e->name, "PCA-200E-%d",index);
      printk(DEV_LABEL "found Fore PCA-200E at 0x%p, IRQ %d\n",
	     pca200e->phys_rambase, pca200e->irq);
      dev = atm_dev_register(DEV_LABEL, &ops, 0); 
      if (!dev) break;
      PCA200E_DEV(dev) = pca200e;
      if (pca200e_init(dev))
	{
	  atm_dev_deregister(dev);
	  break;
	}
      for (i=0; i<ESI_LEN; i++)
	dev->esi[i] = PCA200E_DEV(dev)->mac[i];
      dev->type = "pca200e";
      pca200e->next = pca200e_boards;
      pca200e_boards = pca200e;
      pca200e->atmdev = dev;
      index++;
      pca200e = (struct pca200e_dev *) kmalloc(sizeof(struct pca200e_dev),
					       GFP_KERNEL);
      if (!pca200e) break;
      memset(pca200e, 0, sizeof(struct pca200e_dev));
//break; /* uncomment this to avoid more than one card gets detected */ 
    };
  if (pca200e)
    kfree(pca200e);
  return index;
};


#ifdef MODULE
unsigned int init_module(void)
{
  int i;
  i = (pca200e_detect() == 0);
  return i;
};

void
pca200e_deinit(struct pca200e_dev* pca200e)
{
      switch (pca200e->cardstate)
	{
	case cs_complete:	/* will never be reached */
	case cs_int:		/* free associated interrupt */
	  free_irq(pca200e->irq, pca200e->atmdev);
	
	case cs_init_lb1:
	case cs_init_sb1:

	case cs_init_rx:

	  /*
	   * smlb1_queue and lrgb1_queue have to be freeeeeeed, too.
	   */

	    if (pca200e->smlb1_status_page)
	      free_pages((long) pca200e->smlb1_status_page,
			 (QSIZE_RESUPPLY*32+4095)/4096);
	    if (pca200e->lrgb1_status_page)
	      free_pages((long) pca200e->lrgb1_status_page,
			 (QSIZE_RESUPPLY*32+4095)/4096);
	    if (pca200e->smlb1_rbd_page)
	      free_pages((long) pca200e->smlb1_rbd_page,
			(QSIZE_RESUPPLY*RBD_BLK_SIZE*sizeof(rbd_t)+4095)/4096);
	    if (pca200e->lrgb1_rbd_page)
	      free_pages((long) pca200e->lrgb1_rbd_page,
			(QSIZE_RESUPPLY*RBD_BLK_SIZE*sizeof(rbd_t)+4095)/4096);
	    while(--((unsigned) (pca200e->bufferlist[0])))
	      {
		if (pca200e->bufferlist[(unsigned) (pca200e->bufferlist[0])])
		  {
		    if (pca200e->bufferlist[(unsigned) (pca200e->bufferlist[0])]->buff)
		      kfree(pca200e->bufferlist[(unsigned) (pca200e->bufferlist[0])]->buff);
		    kfree((void*) (pca200e->bufferlist[(unsigned) (pca200e->bufferlist[0])]));
		  };
	      };

	case cs_init_tx:		/* undo all before cs_init_rx */
	  if (pca200e->rx_queue)
	    {
	      qhost_rx_t* tmp1 = pca200e->rx_queue->next;
	      pca200e->rx_queue->next = NULL;
	      while (tmp1)
		{
		  pca200e->rx_queue = tmp1;
		  tmp1 = tmp1->next;
		  kfree((void*) pca200e->rx_queue);
		};
	    };
	  if (pca200e->rx_status_page)
	    free_pages((long) pca200e->rx_status_page,
		       (QSIZE_RX*32+4095)/4096);
	  if (pca200e->rpd_page)
	    free_pages((long) pca200e->rpd_page,
		       (QSIZE_RX*RPD_SIZE+4095)/4096);

	case cs_init_cmd:		/* undo all before cs_init_tx */
	  if (pca200e->tx_queue)
	    {
	      qhost_tx_t* tmp1 = pca200e->tx_queue->next;
	      pca200e->tx_queue->next = NULL;
	      while (tmp1)
		{
		  pca200e->tx_queue = tmp1;
		  tmp1 = tmp1->next;
		  kfree((void*) pca200e->tx_queue);
		};
	    };
	  if (pca200e->tx_status_page)
	    free_pages((long) pca200e->tx_status_page,
		       (QSIZE_TX*32+4095)/4096);
	  if (pca200e->tpd_page)
	    free_pages((long) pca200e->tpd_page,
		       (QSIZE_TX*TPD_SIZE+4095)/4096);

	case cs_init:		/* undo all before cs_init_cmd */
	  if (pca200e->cmd_queue)
	    {
	      qhost_cmd_t* tmp1 = pca200e->cmd_queue->next;
	      pca200e->cmd_queue->next = NULL;
	      while (tmp1)
		{
		  pca200e->cmd_queue = tmp1;
		  tmp1 = tmp1->next;
		  kfree((void*) pca200e->cmd_queue);
		};
	    };
	  if (pca200e->cmd_status_page)
	    free_pages((long) pca200e->cmd_status_page,
		       (QSIZE_CMD*32+4095)/4096);

	case cs_startfirm:	/* reset the card */
	  pca200e_reset(pca200e);

	case cs_loadfirm:	/* invalidate firmware */
	  writel(0, pca200e->virt_rambase
		 +((struct firmheader*) pca200e_firmware)->startoffset); 
	case cs_reset:

	case cs_map:		/* release shared memory mapping */
	  iounmap(pca200e->virt_rambase);

	case cs_blank:
	};
};

void
cleanup_module(void)
{
  struct pca200e_dev* tmp1;
  while (pca200e_boards)
    {
      atm_dev_deregister(pca200e_boards->atmdev);
      pca200e_deinit(pca200e_boards);
      tmp1 = pca200e_boards;
      pca200e_boards = pca200e_boards->next;
      kfree(tmp1);
    };
  printk("pca200e removed\n");
};

#endif



